from pathlib import Path
from itertools import product

import pytest

import prob_conf_mat as pcm
from prob_conf_mat.metrics import METRIC_REGISTRY, AVERAGING_REGISTRY, get_metric
from prob_conf_mat.metrics.abc import AveragedMetric, Metric, RootMetric
from prob_conf_mat.metrics.interface import _parse_kwargs
from prob_conf_mat.io import load_csv


def test_get_metric_values() -> None:
    all_metrics_avgs_combinations = [
        v.aliases[0] for v in METRIC_REGISTRY.values() if v.is_multiclass
    ] + [
        f"{m}@{a}"
        for m, a in product(
            [v.aliases[0] for v in METRIC_REGISTRY.values() if not v.is_multiclass],
            [v.aliases[0] for v in AVERAGING_REGISTRY.values()],
        )
    ]

    study = pcm.Study(
        seed=0,
        num_samples=10000,
        ci_probability=0.95,
    )

    # Add a bunch of experiments
    conf_mat_paths = Path(
        "./documentation/Getting Started/mnist_digits",
    )
    for file_path in sorted(conf_mat_paths.glob("*.csv")):
        # Split the file name to recover the model and fold
        model, fold = file_path.stem.split("_")

        # Load in the confusion matrix using the utility function
        confusion_matrix = load_csv(location=file_path)

        # Add the experiment to the study
        study.add_experiment(
            experiment_name=f"{model}/fold_{fold}",
            confusion_matrix=confusion_matrix,
            prevalence_prior=0,
            confusion_prior=0,
        )

    for metric_str in all_metrics_avgs_combinations:
        # Add a bucnh of metrics
        study.add_metric(metric=metric_str, aggregation="fe_gaussian")

    for metric_str in all_metrics_avgs_combinations:
        study.get_metric_samples(
            metric=metric_str,
            experiment_name="mlp/aggregated",
            sampling_method="posterior",
        )


class TestInterface:
    def test_implemented_metrics(self):
        all_metrics_avgs_combinations = [
            (v.aliases[0], v, None) for v in METRIC_REGISTRY.values()
        ] + [
            (f"{metric_name}@{avg_name}", metric, avg_method)
            for (metric_name, metric), (avg_name, avg_method) in product(
                [
                    (v.aliases[0], v)
                    for v in METRIC_REGISTRY.values()
                    if not v.is_multiclass
                ],
                [(v.aliases[0], v) for v in AVERAGING_REGISTRY.values()],
            )
        ]

        for metric_str, metric, averaging_method in all_metrics_avgs_combinations:
            fetched_metric = get_metric(metric_str)

            # The metric must be an instance of the generating metric
            if isinstance(fetched_metric, Metric):
                assert isinstance(fetched_metric, metric), (fetched_metric, metric)

            # If an averaged metric:
            #   - the base metric must be an instance of the generating metric
            #   - the base averaging method must be an instance of the generating averaging method
            elif isinstance(fetched_metric, AveragedMetric):
                assert isinstance(fetched_metric.base_metric, metric), (
                    fetched_metric.base_metric,
                    metric,
                )

                assert isinstance(fetched_metric.averaging, averaging_method), (
                    fetched_metric.averaging,
                    averaging_method,
                )

    def test_get_metric(self):
        # Test metric syntax string with multiple averaging methods
        with pytest.raises(
            ValueError,
            match="Multiple averaging methods found in metric string",
        ):
            get_metric("foo@bar@baz")

        # Check RootMetric fetch
        assert isinstance(get_metric("norm_confusion_matrix"), RootMetric)

        # Test unknown metric
        with pytest.raises(
            ValueError,
            match="Metric alias must be registered.",
        ):
            get_metric("foobarbaz")

        # Test applying averaging metric to an already multiclass metric
        with pytest.raises(
            ValueError,
            match="Metric is already multivariate and does",
        ):
            get_metric("acc@macro")

        # Test unknown averaging method
        with pytest.raises(
            ValueError,
            match="Averaging alias must be registered.",
        ):
            get_metric("f1@foobarbaz")

    def test_kwargs_parsing(self):
        # Test the conversion of kwargs dict
        assert _parse_kwargs({"foo": "None"})["foo"] is None
        assert (out := _parse_kwargs({"foo": "0"})["foo"]) == 0 and isinstance(out, int)
        assert (out := _parse_kwargs({"foo": "1.5"})["foo"]) == 1.5 and isinstance(
            out,
            float,
        )
        assert isinstance(
            (out := _parse_kwargs({"foo": "[0.5]"})["foo"]),
            list,
        ) and isinstance(out[0], float)
        assert isinstance(
            _parse_kwargs({"foo": "This is a string"})["foo"],
            str,
        )
        assert isinstance(
            _parse_kwargs({"foo": "True"})["foo"],
            bool,
        )

        # Test finding kwargs in metric syntax string
        metric: pcm.metrics._metrics.BalancedAccuracy = get_metric("ba+adjusted=True")
        assert metric.adjusted == True

        metric: pcm.metrics._metrics.AveragedMetric = get_metric(
            "f1@binary+positive_class=0",
        )
        assert metric.averaging.positive_class == 0

        # Test finding unterminated kwargs in metric syntax string
        with pytest.raises(
            ValueError,
            match="Found potentially unterminated kwarg in",
        ):
            get_metric("ba+adjusted")

        with pytest.raises(
            ValueError,
            match="Found potentially unterminated kwarg in",
        ):
            get_metric("f1@binary+positive_class")



import typing


def fmt(number: float, precision: int = 4, mode: str = "f") -> str:
    # Format as float, falling back to scientific notation if too small
    if mode == "f":
        if number != 0.0 and abs(number) <= 1 * (10 ** (-precision)):
            return f"{number:.{precision}e}"
        else:
            return f"{number:.{precision}f}"

    elif mode == "e":
        return f"{number:.{precision}e}"

    elif mode == "%":
        return f"{number * 100:2.{precision - 2}f}%"

    else:
        raise ValueError(
            f"Parameter mode must be one of 'f', 'e', '%'. Currently: {mode}"
        )


def summary_to_table_row(
    name: str, summary: typing.List[typing.Tuple], precision: int = 4
):
    row = [name]
    row += [
        f"[{fmt(val[0], precision=precision)}, {fmt(val[1], precision=precision)}]"
        if "HDI" in stat
        else val
        for stat, val in summary
    ]

    return row



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import numpy as np
    import jaxtyping as jtyping

from prob_conf_mat.metrics.abc import Averaging
from prob_conf_mat.stats import (
    numpy_batched_arithmetic_mean,
    numpy_batched_convex_combination,
    numpy_batched_geometric_mean,
)


class MacroAverage(Averaging):
    """Computes the arithmetic mean over all classes, also known as macro-averaging."""

    full_name = "Macro Averaging"
    dependencies = ()
    sklearn_equivalent = "macro"
    aliases = ["macro", "macro_average", "mean"]

    def compute_average(
        self,
        metric_values: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        scalar_array = numpy_batched_arithmetic_mean(
            metric_values,
            axis=1,
            keepdims=False,
        )

        return scalar_array


class WeightedAverage(Averaging):
    """Computes the class prevalence weighted mean over all classes, also known as weighted averaging."""

    full_name = "Class Prevalence Weighted Averaging"
    dependencies = ("p_condition",)
    sklearn_equivalent = "weighted"
    aliases = ["weighted", "weighted_average", "micro", "micro_average"]

    def compute_average(
        self,
        metric_values: jtyping.Float[np.ndarray, " num_samples num_classes"],
        p_condition: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        scalar_array = numpy_batched_convex_combination(
            metric_values,
            convex_weights=p_condition,
            axis=1,
            keepdims=False,
        )

        return scalar_array


class SelectPositiveClass(Averaging):
    """Selects only the positive class, also known as binary averaging."""

    full_name = "Select Positive Class"
    dependencies = ()
    sklearn_equivalent = "binary"
    aliases = ["select_positive", "binary", "select"]

    def __init__(self, positive_class: int = -1) -> None:
        super().__init__()

        self.positive_class = positive_class

    def compute_average(
        self,
        metric_values: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        if metric_values.shape[1] < self.positive_class:
            raise IndexError(
                f"Passed metric values contain fewer classes than 'positive_class' argument: `{self.positive_class}`.",  # noqa: E501
            )

        scalar_array = metric_values[:, self.positive_class]

        return scalar_array


class HarmonicMean(Averaging):
    """Computes the harmonic mean over all classes."""

    full_name = "Harmonic Mean Averaging"
    dependencies = ()
    sklearn_equivalent = None
    aliases = ["harmonic", "harm"]

    def compute_average(
        self,
        metric_values: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        scalar_array = numpy_batched_geometric_mean(
            metric_values,
            axis=1,
            keepdims=False,
        )

        return scalar_array


class GeometricMean(Averaging):
    """Computes the geometric mean over all classes."""

    full_name = "Geometric Mean Averaging"
    dependencies = ()
    sklearn_equivalent = None
    aliases = ["geometric", "geom"]

    def compute_average(
        self,
        metric_values: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        scalar_array = numpy_batched_geometric_mean(
            metric_values,
            axis=1,
            keepdims=False,
        )

        return scalar_array



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

    from prob_conf_mat.experiment import Experiment

from functools import partial
from pathlib import Path
from itertools import product

import pytest
import numpy as np
import sklearn
import sklearn.metrics

from prob_conf_mat import Study
from prob_conf_mat.io import (
    load_csv,
    confusion_matrix_to_pred_cond,
)

# ==============================================================================
# Setup all test cases
# ==============================================================================
# Metrics ======================================================================
METRICS_TO_SKLEARN = {
    "acc": sklearn.metrics.accuracy_score,
    "ba": sklearn.metrics.balanced_accuracy_score,
    "ba+adjusted=True": partial(sklearn.metrics.balanced_accuracy_score, adjusted=True),
    "cohen_kappa": sklearn.metrics.cohen_kappa_score,
    "mcc": sklearn.metrics.matthews_corrcoef,
    "f1": partial(sklearn.metrics.f1_score, average=None),
    "jaccard": partial(sklearn.metrics.jaccard_score, average=None),
    "plr@binary+positive_class=0": lambda y_true,
    y_pred: sklearn.metrics.class_likelihood_ratios(
        y_true=y_true,
        y_pred=y_pred,
        raise_warning=False,
    )[0],
    "nlr@binary+positive_class=0": lambda y_true,
    y_pred: sklearn.metrics.class_likelihood_ratios(
        y_true=y_true,
        y_pred=y_pred,
        raise_warning=False,
    )[1],
    "dor@binary+positive_class=0": lambda y_true,
    y_pred: sklearn.metrics.class_likelihood_ratios(
        y_true=y_true,
        y_pred=y_pred,
        raise_warning=False,
    )[0]
    / sklearn.metrics.class_likelihood_ratios(
        y_true=y_true,
        y_pred=y_pred,
        raise_warning=False,
    )[1],
    **{
        f"fbeta+beta={beta}": partial(
            sklearn.metrics.fbeta_score,
            average=None,
            beta=beta,
        )
        for beta in [0.0, 0.5, 1.0, 2.0]
    },
}

# Confusion matrices ===========================================================
TEST_CASES_DIR = Path("./tests/data/confusion_matrices")

assert TEST_CASES_DIR.is_dir(), TEST_CASES_DIR

# Their combination ============================================================
all_metrics_to_test = METRICS_TO_SKLEARN.keys()

all_confusion_matrices_to_test = list(TEST_CASES_DIR.glob(pattern="*.csv"))

all_args_to_test = list(
    product(
        all_confusion_matrices_to_test,
        all_metrics_to_test,
    ),
)


# ==============================================================================
# The 'assert' code for pytest
# ==============================================================================
@pytest.mark.parametrize(
    argnames="conf_mat_fp,metric",
    argvalues=all_args_to_test,
)
def test_metric_equivalence(conf_mat_fp, metric) -> None:
    def _get_our_value(
        metric: str,
        study: Study,
    ) -> jtyping.Float[np.ndarray, " num_classes"]:
        metric_result = study.get_metric_samples(
            metric=metric,
            experiment_name="test/test",
            sampling_method="input",
        )

        our_value = metric_result.values[0]

        return our_value

    def _get_sklearn_value(
        metric: str,
        study: Study,
    ) -> jtyping.Float[np.ndarray, " *num_classes"]:
        experiment: Experiment = study["test/test"]  # type: ignore
        conf_mat = experiment.confusion_matrix

        pred_cond = confusion_matrix_to_pred_cond(
            confusion_matrix=conf_mat,
            pred_first=True,
        )

        # Need to binarize the cond_pred array, otherwise sklearn complains
        if "binary+positive_class=0" in metric:
            pred_cond = np.where(pred_cond == 0, 1, 0)

        sklearn_func = METRICS_TO_SKLEARN[metric]

        match metric:
            # For some reason cohen's kappa has a different function signature
            case "cohen_kappa":
                sklearn_value = sklearn_func(y1=pred_cond[:, 0], y2=pred_cond[:, 1])
            case "dor@binary+positive_class=0":
                # Catch zero division error, these aren't compared anyways
                try:
                    sklearn_value = sklearn_func(
                        y_pred=pred_cond[:, 0],
                        y_true=pred_cond[:, 1],
                    )
                except ZeroDivisionError:
                    sklearn_value = np.array(np.nan)
            case _:
                sklearn_value = sklearn_func(
                    y_pred=pred_cond[:, 0],
                    y_true=pred_cond[:, 1],
                )

        return sklearn_value

    study = Study(
        seed=0,
        num_samples=10000,
        ci_probability=0.95,
    )

    study.add_metric(metric=metric)

    conf_mat = load_csv(
        location=conf_mat_fp,
    )

    study.add_experiment(
        experiment_name="test/test",
        confusion_matrix=conf_mat,
        prevalence_prior=0,
        confusion_prior=0,
    )

    our_value = _get_our_value(metric=metric, study=study)

    sklearn_value = _get_sklearn_value(metric=metric, study=study)

    # Only test if either array has finite values
    if np.all(np.isfinite(our_value)) or np.all(np.isfinite(sklearn_value)):
        assert np.allclose(our_value, sklearn_value), (our_value, sklearn_value)
    # If both report non-finite numbers, accept
    else:
        assert True



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

import inspect
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from inspect import signature
from itertools import product

import numpy as np


# Root metrics are always computed, because they're (almost) always needed as
# intermediate variables
_ROOT_METRICS = {
    "norm_confusion_matrix",
    "p_condition",
    "p_pred_given_condition",
    "p_pred",
    "p_condition_given_pred",
}
METRIC_REGISTRY = dict()
AVERAGING_REGISTRY = dict()


@dataclass(frozen=True)
class RootMetric:
    """A metric that all other metrics depend upon, but with no dependencies."""

    name: str

    @property
    def full_name(self) -> str:  # noqa: D102
        return self.name

    @property
    def is_multiclass(self) -> bool:  # noqa: D102
        raise TypeError("Root metrics are not directly interpretable.")

    @property
    def bounds(self) -> tuple[float, float]:  # noqa: D102
        raise TypeError("Root metrics are not directly interpretable.")

    @property
    def dependencies(self) -> typing.Sequence[str]:  # noqa: D102
        return ()

    @property
    def sklearn_equivalent(self) -> str | None:  # noqa: D102
        raise TypeError("Root metrics are not directly interpretable.")

    @property
    def aliases(self) -> list[str]:  # noqa: D102
        return [self.name]


class Metric(metaclass=ABCMeta):
    """The abstract base class for metrics.

    Properties should be implemented as class attributes in derived metrics

    The `compute_metric` method needs to be implemented

    """

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        # Validate =============================================================
        # Make sure that all aliases are unique
        for alias in cls.aliases:  # type: ignore
            if alias in METRIC_REGISTRY:
                raise ValueError(
                    f"Alias '{alias}' not unique. "
                    f"Currently used by metric {METRIC_REGISTRY[alias]}.",
                )

        # Make sure the parameters of the `compute_metric` function are actually
        # the ones listed as dependencies
        parameters = set(signature(cls.compute_metric).parameters.keys()) - {"self"}
        dependencies = set(cls.dependencies)  # type: ignore
        if parameters != dependencies:
            raise TypeError(
                f"The input for the {cls.__name__}'s `compute_metric` method does not match the specified dependencies: {parameters} != {dependencies}",  # noqa: E501
            )

        # Make sure the dependencies actually exist
        if not hasattr(cls, "_validated_dependencies"):
            for dependency in cls.dependencies:
                if dependency in _ROOT_METRICS:
                    continue

        # Register =============================================================
        for alias in cls.aliases:  # type: ignore
            METRIC_REGISTRY[alias] = cls

        cls._kwargs = {
            param.name: param.annotation
            for param in inspect.signature(cls).parameters.values()
        }

    def __init__(self):
        self._instantiation_name: str = ""

    @property
    @abstractmethod
    def full_name(self) -> str:  # type: ignore
        """A human-readable name for this metric."""
        raise NotImplementedError

    full_name: str

    @property
    @abstractmethod
    def is_multiclass(self) -> bool:  # type: ignore
        """Whether or not this metric computes a value for each class individually, or for all classes at once."""  # noqa: E501
        raise NotImplementedError

    is_multiclass: bool

    @property
    @abstractmethod
    def bounds(self) -> tuple[float, float]:  # type: ignore
        """A tuple of the minimum and maximum possible value for this metric to take.

        Can be infinite.
        """
        raise NotImplementedError

    bounds: tuple[float, float]

    @property
    @abstractmethod
    def dependencies(self) -> typing.Sequence[str]:  # type: ignore
        """All metrics upon which this metric depends.

        Used to generate a computation schedule, such that no metric is calculated before its
        dependencies. The dependencies **must** match the `compute_metric` signature.
        This is checked during class definition.
        """
        raise NotImplementedError

    dependencies: typing.Sequence[str]

    @property
    @abstractmethod
    def sklearn_equivalent(self) -> str | None:  # type: ignore
        """The `sklearn` equivalent function, if applicable"""
        raise NotImplementedError

    sklearn_equivalent: str | None

    @property
    @abstractmethod
    def aliases(self) -> typing.Sequence[str]:  # type: ignore
        """A list of all valid aliases for this metric. Can be used when creating metric syntax strings."""
        raise NotImplementedError

    aliases: typing.Sequence[str]

    @abstractmethod
    def compute_metric(self, *args, **kwargs):
        """Computes the metric values from its dependencies."""
        raise NotImplementedError()

    def __call__(
        self,
        *args,
        **kwargs,
    ) -> jtyping.Float[np.ndarray, " num_samples ..."]:
        return self.compute_metric(*args, **kwargs)

    @property
    def _metric_name(self) -> str | None:
        return self.aliases[0]

    @property
    def name(self) -> str:
        if self._instantiation_name != "":
            return self._instantiation_name
        metric_kwargs = "".join(
            [f"+{k}={getattr(self, k)}" for k, _ in self._kwargs.items()],
        )

        return f"{self._metric_name}{metric_kwargs}"

    def __repr__(self) -> str:
        return f"Metric({self.name})"

    def __str__(self) -> str:
        return f"Metric({self.name})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)


class Averaging(metaclass=ABCMeta):
    """The abstract base class for metric averaging.

    Properties should be implemented as class attributes in derived metrics.

    The `compute_average` method needs to be implemented

    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Validate =============================================================
        # Make sure that all aliases are unique
        for alias in cls.aliases:  # type: ignore
            if alias in AVERAGING_REGISTRY:
                raise ValueError(
                    f"Alias '{alias}' not unique. Currently used by averaging method {AVERAGING_REGISTRY[alias]}.",
                )

        for alias in cls.aliases:  # type: ignore
            if alias in METRIC_REGISTRY:
                raise ValueError(
                    f"Alias '{alias}' not unique. Currently used by metric {METRIC_REGISTRY[alias]}.",
                )

        # Register =============================================================
        for alias in cls.aliases:  # type: ignore
            AVERAGING_REGISTRY[alias] = cls

        cls._kwargs = {
            param.name: param.annotation
            for param in inspect.signature(cls).parameters.values()
        }

    def __init__(self):
        self._instantiation_name: str = ""

    @property
    @abstractmethod
    def full_name(self) -> str:  # type: ignore
        raise NotImplementedError

    full_name: str

    @property
    @abstractmethod
    def dependencies(self) -> typing.Sequence[str]:  # type: ignore
        raise NotImplementedError

    dependencies: typing.Sequence[str]

    @property
    @abstractmethod
    def sklearn_equivalent(self) -> str | None:  # type: ignore
        raise NotImplementedError

    sklearn_equivalent: str | None

    @property
    @abstractmethod
    def aliases(self) -> typing.Sequence[str]:  # type: ignore
        raise NotImplementedError

    aliases: typing.Sequence[str]

    @abstractmethod
    def compute_average(self, *args, **kwargs):
        raise NotImplementedError()

    def __call__(
        self,
        metric_vals: jtyping.Float[np.ndarray, " num_samples num_classes"],
        *args,
        **kwargs,
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        return self.compute_average(metric_vals, *args, **kwargs)

    @property
    def _averaging_name(self) -> str:
        return self.aliases[0]

    @property
    def name(self) -> str:
        if self._instantiation_name != "":
            return self._instantiation_name
        kwargs = "".join(
            [f"+{k}={getattr(self, k)}" for k, _ in self._kwargs.items()],
        )

        return self._averaging_name + kwargs

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)


class AveragedMetric(metaclass=ABCMeta):
    """The abstract base class for the composition of any instance of `Metric` with any instance of `Averaging`.

    Args:
        metric (Metric): a binary metric
        averaging (Averaging): an averaging method
    """

    def __init__(self, metric: Metric, averaging: Averaging):
        super().__init__()

        self.base_metric = metric

        if self.base_metric.is_multiclass:
            raise ValueError(
                f"Cannot aggregate a metric ({self.base_metric.name}) that is already multiclass.",  # noqa: E501
            )

        self.averaging = averaging

        self._instantiation_name: str = ""

    @property
    def aliases(self) -> typing.Sequence[str]:
        """A list of all valid aliases for this metric.

        Constructed from the product of the all aliases of the Metric and Averaging methods.

        Can be used when creating metric syntax strings.
        """
        return [
            f"{lhs}@{rhs}"
            for lhs, rhs in product(self.base_metric.aliases, self.averaging.aliases)  # type: ignore
        ]

    @property
    def full_name(self) -> str:
        return f"{self.base_metric.full_name} with {self.averaging.full_name}"

    @property
    def is_multiclass(self) -> typing.Literal[True]:
        """Whether or not this metric computes a value for each class individually, or for all classes at once.

        An AveragedMetric is *always* multiclass.
        """
        return True

    @property
    def bounds(self) -> tuple[float, float]:
        """A tuple of the minimum and maximum possible value for this metric to take. Can be infinite."""
        return self.base_metric.bounds  # type: ignore

    @property
    def dependencies(self) -> typing.Sequence[str]:
        """All metrics upon which this AveragedMetric depends.

        Constructed from the union of all Metric and AveragingMethod dependencies.

        Used to generate a computation schedule, such that no metric is calculated before its dependencies.

        The dependencies **must** match the `compute_metric` signature.

        This is checked during class definition.
        """
        dependencies = (
            *self.base_metric.dependencies,  # type: ignore
            *self.averaging.dependencies,  # type: ignore
        )

        return dependencies

    @property
    def sklearn_equivalent(self) -> str | None:
        """The `sklearn` equivalent function, if applicable"""
        sklearn_equivalent = self.base_metric.sklearn_equivalent
        if self.averaging.sklearn_equivalent is not None:
            sklearn_equivalent = (
                sklearn_equivalent.sklearn_equivalent  # type: ignore
                + f"with average={self.averaging.sklearn_equivalent}"
            )

        return sklearn_equivalent  # type: ignore

    def compute_metric(
        self,
        *args,
        **kwargs,
    ) -> jtyping.Float[np.ndarray, " num_samples ..."]:
        return self.base_metric(*args, **kwargs)

    def compute_average(
        self,
        *args,
        **kwargs,
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        return self.averaging.__call__(*args, **kwargs)

    def __call__(self, **kwargs) -> jtyping.Float[np.ndarray, " num_samples 1"]:
        metric_vals = self.compute_metric(
            **{
                key: value
                for key, value in kwargs.items()
                if key == "samples" or key in self.base_metric.dependencies
            },
        )

        aggregated_metric_vals = self.averaging(
            metric_vals=metric_vals,
            **{
                key: value
                for key, value in kwargs.items()
                if key in self.averaging.dependencies
            },
        )

        return aggregated_metric_vals[:, np.newaxis]

    @property
    def _kwargs(self) -> dict[str, dict[str, typing.Any]]:
        kwargs = {
            "metric": self.base_metric._kwargs,
            "averaging": self.averaging._kwargs,
        }

        return kwargs

    @property
    def name(self) -> str:
        if self._instantiation_name != "":
            return self._instantiation_name
        return f"{self.base_metric.name}@{self.averaging.name}"

    def __repr__(self) -> str:
        return f"AveragedMetric({self.name})"

    def __str__(self) -> str:
        return f"AveragedMetric({self.name})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)



import pytest
import numpy as np


from prob_conf_mat.stats.dirichlet_distribution import dirichlet_prior, dirichlet_sample

ALL_SEEDS = np.random.default_rng(seed=0).integers(low=0, high=2**31 - 1, size=(10,))
ALL_DIMS = [2, 4, 8, 16]


class TestDirichletPrior:
    def test_numeric_strategy(self) -> None:
        # Test float
        dirichlet_prior(strategy=1.0, shape=(2, 2))

        # Test int
        dirichlet_prior(strategy=1, shape=(2, 2))

        # Test bounds
        dirichlet_prior(strategy=0.0, shape=(2, 2))

        dirichlet_prior(strategy=np.nan, shape=(2, 2))

        with pytest.raises(ValueError):
            dirichlet_prior(strategy=-1, shape=(2, 2))

    def test_str(self) -> None:
        # Test valid
        dirichlet_prior(strategy="haldane", shape=(2, 2))

        # Test invalid
        with pytest.raises(ValueError):
            dirichlet_prior(strategy="foobarbaz", shape=(2, 2))

    def test_arraylike(self) -> None:
        # Test valid
        dirichlet_prior(strategy=[[0, 0], [0, 0]], shape=(2, 2))

        dirichlet_prior(strategy=np.zeros((2, 2)), shape=(2, 2))

        # Test invalid shape
        with pytest.raises(ValueError):
            dirichlet_prior(strategy=[[0, 0], [0, 0], [0, 0]], shape=(2, 2))

        # Test invalid object
        class Dummy: ...

        with pytest.raises(ValueError):
            dirichlet_prior(strategy=Dummy, shape=(2, 2))  # type: ignore


class TestDirichletDistribution:
    @pytest.mark.parametrize("seed", argvalues=ALL_SEEDS)
    @pytest.mark.parametrize("dim", argvalues=ALL_DIMS)
    def test_distribution_equivalence_1d(self, seed: int, dim: int):
        rng = np.random.default_rng(seed=seed)

        alphas = rng.integers(low=1, high=10, size=(dim,))

        rng = np.random.default_rng(seed=seed)

        np_dirichlet_samples = rng.dirichlet(alpha=alphas, size=10000)

        rng = np.random.default_rng(seed=seed)

        our_dirichlet_samples = dirichlet_sample(
            rng=rng,  # type: ignore
            alphas=alphas,
            num_samples=10000,
        )

        assert np.allclose(
            np_dirichlet_samples,
            our_dirichlet_samples,
        ), "Distributions do not match"

    @pytest.mark.parametrize("seed", argvalues=ALL_SEEDS)
    @pytest.mark.parametrize("dim", argvalues=ALL_DIMS)
    def test_distribution_equivalence_2d(self, seed: int, dim: int) -> None:
        num_samples = 10000

        rng = np.random.default_rng(seed=seed)

        alphas = rng.integers(low=1, high=10, size=(dim, dim))

        rng = np.random.default_rng(seed=seed)

        np_dirichlet_samples = np.stack(
            arrays=[
                rng.dirichlet(alpha=alphas_, size=num_samples) for alphas_ in alphas
            ],
            axis=1,
        )

        rng = np.random.default_rng(seed=seed)

        our_dirichlet_samples = dirichlet_sample(
            rng=rng,  # type: ignore
            alphas=alphas,
            num_samples=num_samples,
        )

        #! Means of the different distributions must be within 1% of each other
        mean_geom_mean_relative_error = np.abs(
            np.exp(
                np.log(np.mean(our_dirichlet_samples.reshape(num_samples, -1), axis=0))
                - np.log(
                    np.mean(np_dirichlet_samples.reshape(num_samples, -1), axis=0),
                ),
            )
            - 1,
        )

        assert np.mean(mean_geom_mean_relative_error) < 0.01, (
            f"Means of distributions do not match: {np.mean(mean_geom_mean_relative_error)}"
        )

        #! Variances of the different distributions must be within 5% of each other
        var_geom_mean_relative_error = np.abs(
            np.exp(
                np.log(np.var(our_dirichlet_samples.reshape(num_samples, -1), axis=0))
                - np.log(
                    np.var(np_dirichlet_samples.reshape(num_samples, -1), axis=0),
                ),
            )
            - 1,
        )

        assert np.mean(var_geom_mean_relative_error) < 0.05, (
            f"Variances of distributions do not match: {np.mean(var_geom_mean_relative_error)}"
        )



#! The order matters, do not change
from .abc import (
    _ROOT_METRICS,
    METRIC_REGISTRY,
    AVERAGING_REGISTRY,
    RootMetric,
    Metric,
    AveragedMetric,
)
from ._metrics import *
from .averaging import *
from .interface import get_metric
from .collection import MetricCollection

# Check that all metrics have valid dependencies
for metric in METRIC_REGISTRY:
    for dependency in METRIC_REGISTRY[metric].dependencies:
        if dependency in _ROOT_METRICS:
            continue

        try:
            get_metric(dependency)
        except Exception as e:
            raise KeyError(
                f"Dependency `{dependency}` of `{metric}` not valid because: {e}"  # noqa: E501
            )

for aggregation in AVERAGING_REGISTRY:
    for dependency in AVERAGING_REGISTRY[aggregation].dependencies:
        if dependency in _ROOT_METRICS:
            continue

        try:
            get_metric(dependency)
        except Exception as e:
            raise KeyError(
                f"Dependency `{dependency}` of `{aggregation}` not valid because: {e}"  # noqa: E501
            )



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from prob_conf_mat.utils.typing import MetricLike

import re
from functools import cache

from prob_conf_mat.metrics.abc import (
    _ROOT_METRICS,
    METRIC_REGISTRY,
    AVERAGING_REGISTRY,
    AveragedMetric,
    RootMetric,
)

RESERVED_CHARACTERS = {
    "@",  # Denotes the boundary between describing the metric and its averaging method
    "+",  # Denotes a kwarg key
    "=",  # Denotes a kwarg value
}
# TODO: document these regex strings
# Reminder: always document regex immediately after writing down...
NAME_REGEX = re.compile(r"([^\+\@\=]+)[\+\@\=]?")
ARGUMENT_REGEX = re.compile(r"\+([^\+\@\=]+)\=([^\+\@\=]+)[^\+\@]?")
UNTERMINATED_ARGUMENT_REGEX = re.compile(r"\+([^\+\@\=]+)")


def _parse_kwargs(kwargs: dict[str, str]) -> dict[str, typing.Any]:
    """Parse any passed kwargs."""
    for k, v in kwargs.items():
        # First try to eval
        try:
            val = eval(v)

            kwargs[k] = val

        # If it fails, assume a string
        except:  # noqa: E722
            continue

    return kwargs


@cache
def get_metric(syntax_string: str) -> MetricLike:
    """Takes a metric syntax string and returns a metric class instance, potentially with included averaging.

    Args:
        syntax_string (str): a valid metric syntax string

    Returns:
        callable: a metric class instance
    """  # noqa: E501
    # Split on the averaging
    syntax_components = syntax_string.split("@")

    if len(syntax_components) > 2:
        raise ValueError(
            f"Multiple averaging methods found in metric string `{syntax_string}`. Make sure to include only one `@` character",  # noqa: E501
        )

    # Parse the metric name ====================================================
    metric_string = syntax_components[0]

    metric_name = NAME_REGEX.findall(metric_string)[0]

    if metric_name in METRIC_REGISTRY:
        metric_class = METRIC_REGISTRY[metric_name]

    elif metric_name in _ROOT_METRICS:
        return RootMetric(name=metric_name)

    else:
        raise ValueError(
            f"Metric alias must be registered. "
            f"Currently: {metric_name}. "
            f"Must be one of {set(METRIC_REGISTRY.keys())}",
        )

    # Parse and pass the kwargs for the metric function ========================
    metric_kwargs = ARGUMENT_REGEX.findall(metric_string)
    metric_kwargs = dict(metric_kwargs)
    metric_kwargs = _parse_kwargs(metric_kwargs)

    # Check if there are more started kwargs than kwargs with values
    # If so, assume an unterminated kwargs expression
    unterminated_kwargs = UNTERMINATED_ARGUMENT_REGEX.findall(metric_string)
    if len(unterminated_kwargs) != len(metric_kwargs):
        raise ValueError(
            f"Found potentially unterminated kwarg in: {metric_string}. "
            "Make sure kwargs are written as '+foo=bar'",
        )

    metric_instance = metric_class(**metric_kwargs)

    metric_instance._instantiation_name = metric_string

    # Parse the averaging name ===============================================
    if len(syntax_components) == 2:
        averaging_string = syntax_components[1]

        if metric_instance.is_multiclass:
            raise ValueError(
                "Metric is already multivariate and does not need to be averaged. Please remove the `@` specification",  # noqa: E501
            )

        averaging_name = NAME_REGEX.findall(averaging_string)[0]

        try:
            averaging_class = AVERAGING_REGISTRY[averaging_name]
        except KeyError:
            raise ValueError(
                f"Averaging alias must be registered. "
                f"Currently: {averaging_name}. "
                f"Must be one of {set(AVERAGING_REGISTRY.keys())}",
            )

        # Parse and pass the kwargs for the metric function ========================
        averaging_kwargs = ARGUMENT_REGEX.findall(averaging_string)
        averaging_kwargs = dict(averaging_kwargs)
        averaging_kwargs = _parse_kwargs(averaging_kwargs)

        # Check if there are more started kwargs than kwargs with values
        # If so, assume an unterminated kwargs expression
        unterminated_kwargs = UNTERMINATED_ARGUMENT_REGEX.findall(averaging_string)
        if len(unterminated_kwargs) != len(averaging_kwargs):
            raise ValueError(
                f"Found potentially unterminated kwarg in: {averaging_string}. "
                "Make sure kwargs are written as '+foo=bar'",
            )

        averaging_instance = averaging_class(**averaging_kwargs)

        # Compose the metric & averaging function ============================
        composed_metric_instance = AveragedMetric(
            metric=metric_instance,
            averaging=averaging_instance,
        )

        composed_metric_instance._instantiation_name = syntax_string  # type: ignore

        return composed_metric_instance

    return metric_instance



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

    from prob_conf_mat.utils import RNG, MetricLike

from dataclasses import dataclass
from collections import OrderedDict
from enum import StrEnum

import numpy as np

from prob_conf_mat.metrics import RootMetric, MetricCollection
from prob_conf_mat.stats import dirichlet_sample, dirichlet_prior


class SamplingMethod(StrEnum):
    """All implemented confusion matrix sampling methods."""

    POSTERIOR = "posterior"
    PRIOR = "prior"
    RANDOM = "random"
    INPUT = "input"


@dataclass(frozen=True)
class ExperimentResult:
    """A simple wrapper class for results from an experiment.

    Essentially just combines the metadata of an Experiment with a Metric, and stores its output.

    Args:
        experiment (Experiment): the experiment which produced these results
        metric (MetricLike): the metric instance that produced these results
        values (jtyping.Float[np.ndarray, " num_samples #num_classes"]): the actual produced values
    """  # noqa: E501

    experiment: Experiment
    metric: MetricLike

    values: jtyping.Float[np.ndarray, " num_samples #num_classes"]

    @property
    def is_multiclass(self) -> bool:
        """Whether the metric that produced this result is binary or multiclass."""
        return self.metric.is_multiclass

    @property
    def bounds(self) -> tuple[float]:
        """The minimum and maximum possible values of the metric that produced this result."""
        return self.metric.bounds  # type: ignore

    @property
    def num_classes(self) -> int:
        """The number of classes in the experiment that poduced this result."""
        return self.experiment.num_classes

    @property
    def num_samples(self) -> int:
        """How many confusion matrices were sampled."""
        return self.values.shape[0]

    def __repr__(self) -> str:  # noqa: D105
        return f"ExperimentResult(experiment={self.experiment}, metric={self.metric})"

    def __str__(self) -> str:  # noqa: D105
        return f"ExperimentResult(experiment={self.experiment}, metric={self.metric})"


class Experiment:
    """A single experiment, characterized by a confusion matrix.

    It is responsible for generating synthetic confusion matrices, and computing metrics.

    It is typically part of an ExperimentGroup, which in turn is part of a Study. The Study
    and ExperimentGroup are responsible for passing properly validated initialization parameters.

    Args:
        name (str): the name of this experiment
        rng (RNG): the RNG used to control randomness
        confusion_matrix (jtyping.Int[np.ndarray, 'num_classes num_classes']): the confusion matrix
            for this experiment.
        prevalence_prior (str | float | jtyping.Float[np.typing.ArrayLike, 'num_classes'], optional)
            : the prior over the prevalence counts for this experiments.
            Defaults to 0, Haldane's prior.
        confusion_prior (str | float | jtyping.Float[np.typing.ArrayLike, 'num_classes num_classes']
            , optional): the prior over the confusion counts for this experiments
            Defaults to 0, Haldane's prior.

    """

    def __init__(
        self,
        name: str,
        rng: RNG,
        confusion_matrix: jtyping.Int[np.ndarray, " num_classes num_classes"],
        prevalence_prior: str
        | float
        | jtyping.Float[np.typing.ArrayLike, " num_classes"] = 0,
        confusion_prior: str
        | float
        | jtyping.Float[np.typing.ArrayLike, " num_classes num_classes"] = 0,
    ) -> None:
        self.name = name

        # Argument Validation ==================================================
        # Load the confusion matrix
        # Assume it already has been validated
        self.confusion_matrix = confusion_matrix

        # The prior strategy used for defining the Dirichlet prior counts
        self._init_prevalence_prior = prevalence_prior
        self.prevalence_prior = dirichlet_prior(
            strategy=prevalence_prior,
            shape=(self.num_classes,),
        )

        self._init_confusion_prior = confusion_prior
        self.confusion_prior = dirichlet_prior(
            strategy=confusion_prior,
            shape=(self.num_classes, self.num_classes),
        )

        # The RNG
        self.rng = rng

    @property
    def num_classes(self) -> int:
        """The number of classes in this experiment."""
        return self.confusion_matrix.shape[0]

    @property
    def num_predictions(self) -> int:
        """The total number of predictions in the observed confusion matrix."""
        return np.sum(self.confusion_matrix)

    def _wrap_sample_result(
        self,
        norm_confusion_matrix: jtyping.Float[
            np.ndarray,
            " num_samples num_classes num_classes",
        ],
        p_condition: jtyping.Float[np.ndarray, " num_samples num_classes"],
        p_pred_given_condition: jtyping.Float[
            np.ndarray,
            " num_samples num_classes num_classes",
        ],
        p_pred: jtyping.Float[np.ndarray, " num_samples num_classes"],
        p_condition_given_pred: jtyping.Float[
            np.ndarray,
            " num_samples num_classes num_classes",
        ],
    ) -> dict[MetricLike, ExperimentResult]:
        experiment_sample_result: dict[MetricLike, ExperimentResult] = {
            RootMetric("norm_confusion_matrix"): ExperimentResult(
                experiment=self,
                metric=RootMetric("norm_confusion_matrix"),
                values=norm_confusion_matrix,
            ),
            RootMetric("p_condition"): ExperimentResult(
                experiment=self,
                metric=RootMetric("p_condition"),
                values=p_condition,
            ),
            RootMetric("p_pred_given_condition"): ExperimentResult(
                experiment=self,
                metric=RootMetric("p_pred_given_condition"),
                values=p_pred_given_condition,
            ),
            RootMetric("p_pred"): ExperimentResult(
                experiment=self,
                metric=RootMetric("p_pred"),
                values=p_pred,
            ),
            RootMetric("p_condition_given_pred"): ExperimentResult(
                experiment=self,
                metric=RootMetric("p_condition_given_pred"),
                values=p_condition_given_pred,
            ),
        }

        return experiment_sample_result

    def _sample(
        self,
        condition_counts: jtyping.Float[np.ndarray, " num_classes"],
        confusion_matrix: jtyping.Float[np.ndarray, " num_classes num_classes"],
        num_samples: int,
    ) -> dict[MetricLike, ExperimentResult]:
        p_condition = dirichlet_sample(
            rng=self.rng,  # type: ignore
            alphas=condition_counts,
            num_samples=num_samples,
        )

        p_pred_given_condition = dirichlet_sample(
            rng=self.rng,  # type: ignore
            alphas=confusion_matrix,
            num_samples=num_samples,
        )

        norm_confusion_matrix = p_pred_given_condition * p_condition[:, :, np.newaxis]

        p_pred = norm_confusion_matrix.sum(axis=1)

        p_condition_given_pred = norm_confusion_matrix / p_pred[:, np.newaxis, :]

        output = self._wrap_sample_result(
            norm_confusion_matrix=norm_confusion_matrix,
            p_condition=p_condition,
            p_pred_given_condition=p_pred_given_condition,
            p_pred=p_pred,
            p_condition_given_pred=p_condition_given_pred,
        )

        return output

    def sample_input(
        self,
    ) -> dict[MetricLike, ExperimentResult]:
        """For debug purposes: uses the input confusion matrix as the samples.

        Essentially just adds a batch dimension to the existing confusion matrix.
        """
        confusion_matrix = self.confusion_matrix[np.newaxis, :, :]

        norm_confusion_matrix = confusion_matrix / confusion_matrix.sum()

        condition_counts = confusion_matrix.sum(axis=2)
        p_condition = condition_counts / condition_counts.sum()

        p_pred_given_condition = norm_confusion_matrix / p_condition[:, np.newaxis, :]

        prediction_counts = confusion_matrix.sum(axis=1)
        p_pred = prediction_counts / prediction_counts.sum()

        p_condition_given_pred = norm_confusion_matrix / p_pred[:, :, np.newaxis]

        output = self._wrap_sample_result(
            norm_confusion_matrix=norm_confusion_matrix,
            p_condition=p_condition,
            p_pred_given_condition=p_pred_given_condition,
            p_pred=p_pred,
            p_condition_given_pred=p_condition_given_pred,
        )

        return output

    def sample_prior(
        self,
        num_samples: int,
    ) -> dict[MetricLike, ExperimentResult]:
        """Sample confusion matrices from the prior.

        Args:
            num_samples (int): the number of synthetic confusion matrices to sample.

        Returns:
            dict[MetricLike, ExperimentResult]: _description_
        """
        return self._sample(
            num_samples=num_samples,
            condition_counts=self.prevalence_prior,
            confusion_matrix=self.confusion_prior,
        )

    def sample_posterior(
        self,
        num_samples: int,
    ) -> dict[MetricLike, ExperimentResult]:
        """Sample confusion matrices from the posterior.

        Args:
            num_samples (int): the number of synthetic confusion matrices to sample.

        Returns:
            dict[MetricLike, ExperimentResult]: _description_
        """
        condition_counts = self.confusion_matrix.sum(axis=1)
        posterior_condition_counts = self.prevalence_prior + condition_counts

        posterior_pred_given_condtion_counts = (
            self.confusion_prior + self.confusion_matrix
        )

        return self._sample(
            num_samples=num_samples,
            condition_counts=posterior_condition_counts,
            confusion_matrix=posterior_pred_given_condtion_counts,
        )

    def sample_random_model(
        self,
        num_samples: int,
    ) -> dict[MetricLike, ExperimentResult]:
        """Sample from the randomly initialized model distribution.

        It uses the class prevalence from the data, but a random confusion matrix.
        Thus, this should model a random classifier on the used dataset, accounting for class
        imbalance.

        Args:
            num_samples (int): _description_
        """
        condition_counts = self.confusion_matrix.sum(axis=1)
        posterior_condition_counts = self.prevalence_prior + condition_counts

        posterior_pred_given_condtion_counts = (
            self.confusion_prior + self.confusion_matrix
        )

        # Averages over the rows
        random_pred_given_condition_counts = np.broadcast_to(
            np.mean(
                posterior_pred_given_condtion_counts,
                axis=-1,
                keepdims=True,
            ),
            (self.num_classes, self.num_classes),
        )

        return self._sample(
            num_samples=num_samples,
            condition_counts=posterior_condition_counts,
            confusion_matrix=random_pred_given_condition_counts,
        )

    def sample(
        self,
        sampling_method: SamplingMethod,
        num_samples: int,
    ) -> dict[MetricLike, ExperimentResult]:
        """Sample synthetic confusion matrices for this experiment.

        Args:
            sampling_method (SamplingMethod): the sampling method used to generate the metric
                values. Must a member of the SamplingMethod enum
            num_samples (int): the number of synthetic confusion matrices to sample

        Returns:
            dict[MetricLike, ExperimentResult]: a dictionary of RootMetric
                instances
        """
        # print(f"Calling sample: {self.name}, {sampling_method}, {num_samples}")

        root_metrics = dict()

        match sampling_method:
            case SamplingMethod.POSTERIOR.value:
                root_metrics.update(self.sample_posterior(num_samples=num_samples))
            case SamplingMethod.PRIOR.value:
                root_metrics.update(self.sample_prior(num_samples=num_samples))
            case SamplingMethod.RANDOM.value:
                root_metrics.update(self.sample_random_model(num_samples=num_samples))
            case SamplingMethod.INPUT.value:
                root_metrics.update(self.sample_input())
            case _:
                raise ValueError(
                    (
                        f"Parameter `sampling_method` must be one of "
                        f"{tuple(sm.value for sm in SamplingMethod)}. Currently: {sampling_method}"
                    ),
                )

        return root_metrics

    def sample_metrics(
        self,
        metrics: MetricCollection,
        sampling_method: SamplingMethod,
        num_samples: int,
    ) -> dict[MetricLike, ExperimentResult]:
        """Computes metrics over the synthetic confusion matrices.

        Args:
            metrics (MetricCollection): the metrics needed to be computed on the synthetic confusion
                matrices
            sampling_method (SamplingMethod): the sampling method used to generate the metric
                values. Must a member of the SamplingMethod enum
            num_samples (int): the number of synthetic confusion matrices to sample

        Returns:
            dict[MetricLike, ExperimentResult]: a mapping from metric instance to
                an `ExperimentResult` instance
        """
        # Get the topological ordering of the metrics, such that no metric is computed before
        # its dependencies are
        metric_compute_order = metrics.get_compute_order()

        # First have the experiment generate synthetic confusion matrices and needed RootMetrics
        # dict[RootMetric, ExperimentResult]
        intermediate_results: dict[MetricLike, ExperimentResult] = self.sample(
            sampling_method=sampling_method,
            num_samples=num_samples,
        )

        # Go through all metrics and dependencies in order
        for metric in metric_compute_order:
            # Root metrics have no dependency per-definition
            # and are already computed automatically
            if isinstance(metric, RootMetric):
                continue

            # Filter out all the dependencies for the current metric
            # Since we allow each metric to define it's own dependencies by name (or alias)
            # We have to be a little lenient with how we look these up
            dependencies: dict[str, np.ndarray] = dict()
            for dependency_name in metric.dependencies:
                dependency = metric_compute_order[dependency_name]

                dependencies[dependency_name] = intermediate_results[dependency].values

            # Compute the current metric and add it to the dict
            metric_values = metric(**dependencies)

            # Add the metric values to the intermediate stats dictionary
            intermediate_results[metric] = ExperimentResult(
                experiment=self,
                metric=metric,
                values=metric_values,
            )

        results = OrderedDict(
            [
                (metric, intermediate_results[metric])
                for metric in metrics.get_insert_order()
            ],
        )

        return results

    def __repr__(self) -> str:  # noqa: D105
        return f"Experiment({self.name})"

    def __str__(self) -> str:  # noqa: D105
        return f"Experiment({self.name})"
